@@ -19,6 +19,7 @@
 #include "cron_job.h"
 
 #include "access/genam.h"
+#include "access/hash.h"
 #include "access/heapam.h"
 #include "access/htup_details.h"
 #include "access/skey.h"
@@ -33,6 +34,7 @@
 #include "postmaster/postmaster.h"
 #include "pgstat.h"
 #include "storage/lock.h"
+#include "utils/acl.h"
 #include "utils/builtins.h"
 #include "utils/fmgroids.h"
 #include "utils/inval.h"
@@ -46,30 +48,43 @@
 #include "utils/varlena.h"
 #endif
 
+#include "executor/spi.h"
+#include "catalog/pg_type.h"
+
+#if (PG_VERSION_NUM < 120000)
+#define table_open(r, l) heap_open(r, l)
+#define table_close(r, l) heap_close(r, l)
+#endif
 
 #define EXTENSION_NAME "pg_cron"
 #define CRON_SCHEMA_NAME "cron"
 #define JOBS_TABLE_NAME "job"
 #define JOB_ID_INDEX_NAME "job_pkey"
 #define JOB_ID_SEQUENCE_NAME "cron.jobid_seq"
+#define JOB_RUN_DETAILS_TABLE_NAME "job_run_details"
+#define RUN_ID_SEQUENCE_NAME "cron.runid_seq"
 
 
 /* forward declarations */
 static HTAB * CreateCronJobHash(void);
 
-static int64 NextJobId(void);
+static int64 ScheduleCronJob(Name jobName, char *schedule, char *command);
 static Oid CronExtensionOwner(void);
+static void EnsureDeletePermission(Relation cronJobsTable, HeapTuple heapTuple);
 static void InvalidateJobCacheCallback(Datum argument, Oid relationId);
 static void InvalidateJobCache(void);
 static Oid CronJobRelationId(void);
 
 static CronJob * TupleToCronJob(TupleDesc tupleDescriptor, HeapTuple heapTuple);
 static bool PgCronHasBeenLoaded(void);
+static bool JobRunDetailsTableExists(void);
 
 
 /* SQL-callable functions */
 PG_FUNCTION_INFO_V1(cron_schedule);
+PG_FUNCTION_INFO_V1(cron_schedule_named);
 PG_FUNCTION_INFO_V1(cron_unschedule);
+PG_FUNCTION_INFO_V1(cron_unschedule_named);
 PG_FUNCTION_INFO_V1(cron_job_cache_invalidate);
 
 
@@ -154,7 +169,7 @@ GetCronJob(int64 jobId)
 
 
 /*
- * cluster_schedule schedules a cron job.
+ * cron_schedule schedules an unnamed cron job.
  */
 Datum
 cron_schedule(PG_FUNCTION_ARGS)
@@ -162,21 +177,57 @@ cron_schedule(PG_FUNCTION_ARGS)
 	text *scheduleText = PG_GETARG_TEXT_P(0);
 	text *commandText = PG_GETARG_TEXT_P(1);
 
+	Name jobName = NULL;
 	char *schedule = text_to_cstring(scheduleText);
 	char *command = text_to_cstring(commandText);
+
+	int64 jobId = ScheduleCronJob(jobName, schedule, command);
+
+	PG_RETURN_INT64(jobId);
+}
+
+
+/*
+ * cron_schedule_named schedules a named cron job
+ */
+Datum
+cron_schedule_named(PG_FUNCTION_ARGS)
+{
+	Name jobName = PG_GETARG_NAME(0);
+	text *scheduleText = PG_GETARG_TEXT_P(1);
+	text *commandText = PG_GETARG_TEXT_P(2);
+
+	char *schedule = text_to_cstring(scheduleText);
+	char *command = text_to_cstring(commandText);
+
+	int64 jobId = ScheduleCronJob(jobName, schedule, command);
+
+	PG_RETURN_INT64(jobId);
+}
+
+
+/*
+ * ScheduleCronJob schedules a cron job with the given name.
+ */
+static int64
+ScheduleCronJob(Name jobName, char *schedule, char *command)
+{
 	entry *parsedSchedule = NULL;
 
 	int64 jobId = 0;
 	Datum jobIdDatum = 0;
 
-	Oid cronSchemaId = InvalidOid;
-	Oid cronJobsRelationId = InvalidOid;
+	StringInfoData querybuf;
+	Oid argTypes[7];
+	Datum argValues[7];
+	int argCount = 0;
 
-	Relation cronJobsTable = NULL;
-	TupleDesc tupleDescriptor = NULL;
-	HeapTuple heapTuple = NULL;
-	Datum values[Natts_cron_job];
-	bool isNulls[Natts_cron_job];
+	Oid savedUserId = InvalidOid;
+	int savedSecurityContext = 0;
+
+	TupleDesc returnedRowDescriptor = NULL;
+	HeapTuple returnedRow = NULL;
+	bool returnedJobIdIsNull = false;
 
 	Oid userId = GetUserId();
 	char *userName = GetUserNameFromId(userId, false);
@@ -190,53 +241,109 @@ cron_schedule(PG_FUNCTION_ARGS)
 
 	free_entry(parsedSchedule);
 
-	/* form new job tuple */
-	memset(values, 0, sizeof(values));
-	memset(isNulls, false, sizeof(isNulls));
+	initStringInfo(&querybuf);
 
-	jobId = NextJobId();
-	jobIdDatum = Int64GetDatum(jobId);
+	appendStringInfo(&querybuf,
+		"insert into %s (schedule, command, nodename, nodeport, database, username",
+		quote_qualified_identifier(CRON_SCHEMA_NAME, JOBS_TABLE_NAME));
 
-	values[Anum_cron_job_jobid - 1] = jobIdDatum;
-	values[Anum_cron_job_schedule - 1] = CStringGetTextDatum(schedule);
-	values[Anum_cron_job_command - 1] = CStringGetTextDatum(command);
-	values[Anum_cron_job_nodename - 1] = CStringGetTextDatum(CronHost);
-	values[Anum_cron_job_nodeport - 1] = Int32GetDatum(PostPortNumber);
-	values[Anum_cron_job_database - 1] = CStringGetTextDatum(CronTableDatabaseName);
-	values[Anum_cron_job_username - 1] = CStringGetTextDatum(userName);
-	values[Anum_cron_job_active - 1] = BoolGetDatum(true);
+	if (jobName != NULL)
+	{
+		appendStringInfo(&querybuf, ", jobname");
+	}
 
-	cronSchemaId = get_namespace_oid(CRON_SCHEMA_NAME, false);
-	cronJobsRelationId = get_relname_relid(JOBS_TABLE_NAME, cronSchemaId);
+	appendStringInfo(&querybuf, ") values ($1, $2, $3, $4, $5, $6");
 
-	/* open jobs relation and insert new tuple */
-	cronJobsTable = heap_open(cronJobsRelationId, RowExclusiveLock);
+	if (jobName != NULL)
+	{
+		appendStringInfo(&querybuf, ", $7) ");
+		appendStringInfo(&querybuf, "on conflict on constraint jobname_username_uniq ");
+		appendStringInfo(&querybuf, "do update set ");
+		appendStringInfo(&querybuf, "schedule = EXCLUDED.schedule, ");
+		appendStringInfo(&querybuf, "command = EXCLUDED.command");
+	}
+	else
+	{
+		appendStringInfo(&querybuf, ")");
+	}
 
-	tupleDescriptor = RelationGetDescr(cronJobsTable);
-	heapTuple = heap_form_tuple(tupleDescriptor, values, isNulls);
+	appendStringInfo(&querybuf, " returning jobid");
 
-#if (PG_VERSION_NUM >= 100000)
-	CatalogTupleInsert(cronJobsTable, heapTuple);
-#else
-	simple_heap_insert(cronJobsTable, heapTuple);
-	CatalogUpdateIndexes(cronJobsTable, heapTuple);
-#endif
-	CommandCounterIncrement();
+	argTypes[0] = TEXTOID;
+	argValues[0] = CStringGetTextDatum(schedule);
+	argCount++;
 
-	/* close relation and invalidate previous cache entry */
-	heap_close(cronJobsTable, NoLock);
+	argTypes[1] = TEXTOID;
+	argValues[1] = CStringGetTextDatum(command);
+	argCount++;
+
+	argTypes[2] = TEXTOID;
+	argValues[2] = CStringGetTextDatum(CronHost);
+	argCount++;
+
+	argTypes[3] = INT4OID;
+	argValues[3] = Int32GetDatum(PostPortNumber);
+	argCount++;
+
+	argTypes[4] = TEXTOID;
+	argValues[4] = CStringGetTextDatum(CronTableDatabaseName);
+	argCount++;
+
+	argTypes[5] = TEXTOID;
+	argValues[5] = CStringGetTextDatum(userName);
+	argCount++;
+
+	if (jobName != NULL)
+	{
+		argTypes[6] = NAMEOID;
+		argValues[6] = NameGetDatum(jobName);
+		argCount++;
+	}
+
+	GetUserIdAndSecContext(&savedUserId, &savedSecurityContext);
+	SetUserIdAndSecContext(CronExtensionOwner(), SECURITY_LOCAL_USERID_CHANGE);
+
+	/* Open SPI context. */
+	if (SPI_connect() != SPI_OK_CONNECT)
+	{
+		elog(ERROR, "SPI_connect failed");
+	}
+
+	if (SPI_execute_with_args(querybuf.data, argCount, argTypes, argValues, NULL,
+							  false, 1) != SPI_OK_INSERT_RETURNING)
+	{
+		elog(ERROR, "SPI_exec failed: %s", querybuf.data);
+	}
+
+	if (SPI_processed <= 0)
+	{
+		elog(ERROR, "query did not return any rows: %s", querybuf.data);
+	}
+
+	returnedRowDescriptor = SPI_tuptable->tupdesc;
+	returnedRow = SPI_tuptable->vals[0];
+
+	jobIdDatum = SPI_getbinval(returnedRow, returnedRowDescriptor, 1,
+							   &returnedJobIdIsNull);
+	jobId = DatumGetInt64(jobIdDatum);
+
+	pfree(querybuf.data);
+
+	SPI_finish();
+
+	SetUserIdAndSecContext(savedUserId, savedSecurityContext);
 
 	InvalidateJobCache();
 
-	PG_RETURN_INT64(jobId);
+	return jobId;
 }
 
 
 /*
- * NextJobId returns a new, unique job ID using the job ID sequence.
+ * NextRunId draws a new run ID from cron.runid_seq.
  */
-static int64
-NextJobId(void)
+int64
+NextRunId(void)
 {
 	text *sequenceName = NULL;
 	Oid sequenceId = InvalidOid;
@@ -249,8 +356,20 @@ NextJobId(void)
 	int64 jobId = 0;
 	bool failOK = true;
 
+	StartTransactionCommand();
+	PushActiveSnapshot(GetTransactionSnapshot());
+
+	if (!JobRunDetailsTableExists())
+	{
+		PopActiveSnapshot();
+		CommitTransactionCommand();
+
+		/* if the job_run_details table is not yet created, the run ID is not used */
+		return 0;
+	}
+
 	/* resolve relationId from passed in schema and relation name */
-	sequenceName = cstring_to_text(JOB_ID_SEQUENCE_NAME);
+	sequenceName = cstring_to_text(RUN_ID_SEQUENCE_NAME);
 	sequenceNameList = textToQualifiedNameList(sequenceName);
 	sequenceVar = makeRangeVarFromNameList(sequenceNameList);
 	sequenceId = RangeVarGetRelid(sequenceVar, NoLock, failOK);
@@ -266,10 +385,12 @@ NextJobId(void)
 
 	jobId = DatumGetInt64(jobIdDatum);
 
+	PopActiveSnapshot();
+	CommitTransactionCommand();
+
 	return jobId;
 }
 
-
 /*
  * CronExtensionOwner returns the name of the user that owns the
  * extension.
@@ -284,7 +405,7 @@ CronExtensionOwner(void)
 	Form_pg_extension extensionForm = NULL;
 	Oid extensionOwner = InvalidOid;
 
-	extensionRelation = heap_open(ExtensionRelationId, AccessShareLock);
+	extensionRelation = table_open(ExtensionRelationId, AccessShareLock);
 
 	ScanKeyInit(&entry[0],
 				Anum_pg_extension_extname,
@@ -305,14 +426,14 @@ CronExtensionOwner(void)
 	extensionOwner = extensionForm->extowner;
 
 	systable_endscan(scanDescriptor);
-	heap_close(extensionRelation, AccessShareLock);
+	table_close(extensionRelation, AccessShareLock);
 
 	return extensionOwner;
 }
 
 
 /*
- * cluster_unschedule removes a cron job.
+ * cron_unschedule removes a cron job.
  */
 Datum
 cron_unschedule(PG_FUNCTION_ARGS)
@@ -327,18 +448,12 @@ cron_unschedule(PG_FUNCTION_ARGS)
 	ScanKeyData scanKey[1];
 	int scanKeyCount = 1;
 	bool indexOK = true;
-	TupleDesc tupleDescriptor = NULL;
 	HeapTuple heapTuple = NULL;
-	bool isNull = false;
-	Oid userId = InvalidOid;
-	char *userName = NULL;
-	Datum ownerNameDatum = 0;
-	char *ownerName = NULL;
 
 	cronSchemaId = get_namespace_oid(CRON_SCHEMA_NAME, false);
 	cronJobIndexId = get_relname_relid(JOB_ID_INDEX_NAME, cronSchemaId);
 
-	cronJobsTable = heap_open(CronJobRelationId(), RowExclusiveLock);
+	cronJobsTable = table_open(CronJobRelationId(), RowExclusiveLock);
 
 	ScanKeyInit(&scanKey[0], Anum_cron_job_jobid,
 				BTEqualStrategyNumber, F_INT8EQ, Int64GetDatum(jobId));
@@ -347,8 +462,6 @@ cron_unschedule(PG_FUNCTION_ARGS)
 										cronJobIndexId, indexOK,
 										NULL, scanKeyCount, scanKey);
 
-	tupleDescriptor = RelationGetDescr(cronJobsTable);
-
 	heapTuple = systable_getnext(scanDescriptor);
 	if (!HeapTupleIsValid(heapTuple))
 	{
@@ -356,13 +469,88 @@ cron_unschedule(PG_FUNCTION_ARGS)
 							   INT64_FORMAT, jobId)));
 	}
 
-	/* check if the current user owns the row */
-	userId = GetUserId();
-	userName = GetUserNameFromId(userId, false);
+	EnsureDeletePermission(cronJobsTable, heapTuple);
 
-	ownerNameDatum = heap_getattr(heapTuple, Anum_cron_job_username,
-								  tupleDescriptor, &isNull);
-	ownerName = TextDatumGetCString(ownerNameDatum);
+	simple_heap_delete(cronJobsTable, &heapTuple->t_self);
+
+	systable_endscan(scanDescriptor);
+	table_close(cronJobsTable, NoLock);
+
+	CommandCounterIncrement();
+	InvalidateJobCache();
+
+	PG_RETURN_BOOL(true);
+}
+
+
+/*
+ * cron_unschedule_named removes a cron job by name.
+ */
+Datum
+cron_unschedule_named(PG_FUNCTION_ARGS)
+{
+	Datum jobNameDatum = PG_GETARG_DATUM(0);
+	Name jobName = DatumGetName(jobNameDatum);
+
+	Oid userId = GetUserId();
+	char *userName = GetUserNameFromId(userId, false);
+	Datum userNameDatum = CStringGetTextDatum(userName);
+
+	Relation cronJobsTable = NULL;
+	SysScanDesc scanDescriptor = NULL;
+	ScanKeyData scanKey[2];
+	int scanKeyCount = 2;
+	bool indexOK = false;
+	HeapTuple heapTuple = NULL;
+
+	cronJobsTable = table_open(CronJobRelationId(), RowExclusiveLock);
+
+	ScanKeyInit(&scanKey[0], Anum_cron_job_jobname,
+				BTEqualStrategyNumber, F_NAMEEQ, jobNameDatum);
+	ScanKeyInit(&scanKey[1], Anum_cron_job_username,
+				BTEqualStrategyNumber, F_TEXTEQ, userNameDatum);
+
+	scanDescriptor = systable_beginscan(cronJobsTable, InvalidOid, indexOK,
+										NULL, scanKeyCount, scanKey);
+
+	heapTuple = systable_getnext(scanDescriptor);
+	if (!HeapTupleIsValid(heapTuple))
+	{
+		ereport(ERROR, (errmsg("could not find valid entry for job '%s'",
+							   NameStr(*jobName))));
+	}
+
+	EnsureDeletePermission(cronJobsTable, heapTuple);
+
+	simple_heap_delete(cronJobsTable, &heapTuple->t_self);
+
+	systable_endscan(scanDescriptor);
+	table_close(cronJobsTable, NoLock);
+
+	CommandCounterIncrement();
+	InvalidateJobCache();
+
+	PG_RETURN_BOOL(true);
+}
+
+
+/*
+ * EnsureDeletePermission throws an error if the current user does
+ * not have permission to delete the given cron.job tuple.
+ */
+static void
+EnsureDeletePermission(Relation cronJobsTable, HeapTuple heapTuple)
+{
+	TupleDesc tupleDescriptor = RelationGetDescr(cronJobsTable);
+
+	/* check if the current user owns the row */
+	Oid userId = GetUserId();
+	char *userName = GetUserNameFromId(userId, false);
+
+	bool isNull = false;
+	Datum ownerNameDatum = heap_getattr(heapTuple, Anum_cron_job_username,
+										tupleDescriptor, &isNull);
+	char *ownerName = TextDatumGetCString(ownerNameDatum);
 	if (pg_strcasecmp(userName, ownerName) != 0)
 	{
 		/* otherwise, allow if the user has DELETE permission */
@@ -379,16 +567,6 @@ cron_unschedule(PG_FUNCTION_ARGS)
 						   get_rel_name(CronJobRelationId()));
 		}
 	}
-
-	simple_heap_delete(cronJobsTable, &heapTuple->t_self);
-
-	systable_endscan(scanDescriptor);
-	heap_close(cronJobsTable, NoLock);
-
-	CommandCounterIncrement();
-	InvalidateJobCache();
-
-	PG_RETURN_BOOL(true);
 }
 
 
@@ -461,7 +639,6 @@ CronJobRelationId(void)
 	return CachedCronJobRelationId;
 }
 
-
 /*
  * LoadCronJobList loads the current list of jobs from the
  * cron.job table and adds each job to the CronJobHash.
@@ -498,7 +675,7 @@ LoadCronJobList(void)
 		return NIL;
 	}
 
-	cronJobTable = heap_open(CronJobRelationId(), AccessShareLock);
+	cronJobTable = table_open(CronJobRelationId(), AccessShareLock);
 
 	scanDescriptor = systable_beginscan(cronJobTable,
 										InvalidOid, false,
@@ -523,7 +700,7 @@ LoadCronJobList(void)
 	}
 
 	systable_endscan(scanDescriptor);
-	heap_close(cronJobTable, AccessShareLock);
+	table_close(cronJobTable, AccessShareLock);
 
 	PopActiveSnapshot();
 	CommitTransactionCommand();
@@ -563,8 +740,6 @@ TupleToCronJob(TupleDesc tupleDescriptor, HeapTuple heapTuple)
 	Datum userName = heap_getattr(heapTuple, Anum_cron_job_username,
 								  tupleDescriptor, &isNull);
 
-	Assert(!HeapTupleHasNulls(heapTuple));
-
 	jobKey = DatumGetInt64(jobId);
 	job = hash_search(CronJobHash, &jobKey, HASH_ENTER, &isPresent);
 
@@ -588,6 +763,21 @@ TupleToCronJob(TupleDesc tupleDescriptor, HeapTuple heapTuple)
 		job->active = true;
 	}
 
+	if (tupleDescriptor->natts >= Anum_cron_job_jobname)
+	{
+		bool isJobNameNull = false;
+		Datum jobName = heap_getattr(heapTuple, Anum_cron_job_jobname,
+									 tupleDescriptor, &isJobNameNull);
+		if (!isJobNameNull)
+		{
+			job->jobName = DatumGetName(jobName);
+		}
+		else
+		{
+			job->jobName = NULL;
+		}
+	}
+
 	parsedSchedule = parse_cron_entry(job->scheduleText);
 	if (parsedSchedule != NULL)
 	{
@@ -644,3 +834,258 @@ PgCronHasBeenLoaded(void)
 
 	return extensionLoaded;
 }
+
+void
+InsertJobRunDetail(int64 runId, int64 *jobId, char *database, char *username, char *command, char *status)
+{
+	StringInfoData querybuf;
+	const int argCount = 6;
+	Oid argTypes[6];
+	Datum argValues[6];
+
+	SetCurrentStatementStartTimestamp();
+	StartTransactionCommand();
+	PushActiveSnapshot(GetTransactionSnapshot());
+
+	if (!PgCronHasBeenLoaded() || RecoveryInProgress() || !JobRunDetailsTableExists())
+	{
+		PopActiveSnapshot();
+		CommitTransactionCommand();
+		return;
+	}
+
+	initStringInfo(&querybuf);
+
+	/* Open SPI context. */
+	if (SPI_connect() != SPI_OK_CONNECT)
+		elog(ERROR, "SPI_connect failed");
+
+
+	appendStringInfo(&querybuf,
+		"insert into %s.%s (jobid, runid, database, username, command, status) values ($1,$2,$3,$4,$5,$6)",
+		CRON_SCHEMA_NAME, JOB_RUN_DETAILS_TABLE_NAME);
+
+	/* jobId */
+	argTypes[0] = INT8OID;
+	argValues[0] = Int64GetDatum(*jobId);
+
+	/* runId */
+	argTypes[1] = INT8OID;
+	argValues[1] = Int64GetDatum(runId);
+
+	/* database */
+	argTypes[2] = TEXTOID;
+	argValues[2] = CStringGetTextDatum(database);
+
+	/* username */
+	argTypes[3] = TEXTOID;
+	argValues[3] = CStringGetTextDatum(username);
+
+	/* command */
+	argTypes[4] = TEXTOID;
+	argValues[4] = CStringGetTextDatum(command);
+
+	/* status */
+	argTypes[5] = TEXTOID;
+	argValues[5] = CStringGetTextDatum(status);
+
+	pgstat_report_activity(STATE_RUNNING, querybuf.data);
+
+	if(SPI_execute_with_args(querybuf.data,
+		argCount, argTypes, argValues, NULL, false, 1) != SPI_OK_INSERT)
+		elog(ERROR, "SPI_exec failed: %s", querybuf.data);
+
+	pfree(querybuf.data);
+
+	SPI_finish();
+	PopActiveSnapshot();
+	CommitTransactionCommand();
+	pgstat_report_activity(STATE_IDLE, NULL);
+}
+
+void
+UpdateJobRunDetail(int64 runId, int32 *job_pid, char *status, char *return_message, TimestampTz *start_time,
+                                                                        TimestampTz *end_time)
+{
+	StringInfoData querybuf;
+	Oid argTypes[6];
+	Datum argValues[6];
+	int i;
+
+	SetCurrentStatementStartTimestamp();
+	StartTransactionCommand();
+	PushActiveSnapshot(GetTransactionSnapshot());
+
+	if (!PgCronHasBeenLoaded() || RecoveryInProgress() || !JobRunDetailsTableExists())
+	{
+		PopActiveSnapshot();
+		CommitTransactionCommand();
+		return;
+	}
+
+	initStringInfo(&querybuf);
+	i = 0;
+
+	/* Open SPI context. */
+	if (SPI_connect() != SPI_OK_CONNECT)
+		elog(ERROR, "SPI_connect failed");
+
+
+	appendStringInfo(&querybuf,
+		"update %s.%s set", CRON_SCHEMA_NAME, JOB_RUN_DETAILS_TABLE_NAME);
+
+
+	/* add the fields to be updated */
+	if (job_pid != NULL) {
+		argTypes[i] = INT4OID;
+		argValues[i] = Int32GetDatum(*job_pid);
+		i++;
+		appendStringInfo(&querybuf, " job_pid = $%d,", i);
+	}
+
+	if (status != NULL)
+	{
+		argTypes[i] = TEXTOID;
+		argValues[i] = CStringGetTextDatum(status);
+		i++;
+
+		appendStringInfo(&querybuf, " status = $%d,", i);
+	}
+
+        if (return_message != NULL)
+	{
+		argTypes[i] = TEXTOID;
+		argValues[i] = CStringGetTextDatum(return_message);
+		i++;
+
+		appendStringInfo(&querybuf, " return_message = $%d,", i);
+	}
+
+        if (start_time != NULL)
+	{
+		argTypes[i] = TIMESTAMPTZOID;
+		argValues[i] = TimestampTzGetDatum(*start_time);
+		i++;
+
+		appendStringInfo(&querybuf, " start_time = $%d,", i);
+	}
+
+        if (end_time != NULL)
+	{
+		argTypes[i] = TIMESTAMPTZOID;
+		argValues[i] = TimestampTzGetDatum(*end_time);
+		i++;
+
+		appendStringInfo(&querybuf, " end_time = $%d,", i);
+	}
+
+	argTypes[i] = INT8OID;
+	argValues[i] = Int64GetDatum(runId);
+	i++;
+
+	/* remove the last comma */
+	querybuf.len--;
+	querybuf.data[querybuf.len] = '\0';
+
+	/* and add the where clause */
+	appendStringInfo(&querybuf, " where runid = $%d", i);
+
+	pgstat_report_activity(STATE_RUNNING, querybuf.data);
+
+	if(SPI_execute_with_args(querybuf.data,
+		i, argTypes, argValues, NULL, false, 1) != SPI_OK_UPDATE)
+		elog(ERROR, "SPI_exec failed: %s", querybuf.data);
+
+	pfree(querybuf.data);
+
+	SPI_finish();
+	PopActiveSnapshot();
+	CommitTransactionCommand();
+	pgstat_report_activity(STATE_IDLE, NULL);
+}
+
+void
+MarkPendingRunsAsFailed(void)
+{
+	StringInfoData querybuf;
+
+	SetCurrentStatementStartTimestamp();
+	StartTransactionCommand();
+	PushActiveSnapshot(GetTransactionSnapshot());
+
+	if (!PgCronHasBeenLoaded() || RecoveryInProgress() || !JobRunDetailsTableExists())
+	{
+		PopActiveSnapshot();
+		CommitTransactionCommand();
+		return;
+	}
+
+	initStringInfo(&querybuf);
+
+	/* Open SPI context. */
+	if (SPI_connect() != SPI_OK_CONNECT)
+		elog(ERROR, "SPI_connect failed");
+
+
+	appendStringInfo(&querybuf,
+		"update %s.%s set status = '%s', return_message = 'server restarted' where status in ('%s','%s')"
+		, CRON_SCHEMA_NAME, JOB_RUN_DETAILS_TABLE_NAME, GetCronStatus(CRON_STATUS_FAILED), GetCronStatus(CRON_STATUS_STARTING), GetCronStatus(CRON_STATUS_RUNNING));
+
+
+	pgstat_report_activity(STATE_RUNNING, querybuf.data);
+
+	if (SPI_exec(querybuf.data, 0) != SPI_OK_UPDATE)
+		elog(ERROR, "SPI_exec failed: %s", querybuf.data);
+
+	pfree(querybuf.data);
+
+	SPI_finish();
+	PopActiveSnapshot();
+	CommitTransactionCommand();
+	pgstat_report_activity(STATE_IDLE, NULL);
+}
+
+char *
+GetCronStatus(CronStatus cronstatus)
+{
+	char *statusDesc = "unknown status";
+
+	switch (cronstatus)
+	{
+	case CRON_STATUS_STARTING:
+		statusDesc = "starting";
+		break;
+	case CRON_STATUS_RUNNING:
+		statusDesc = "running";
+		break;
+	case CRON_STATUS_SENDING:
+		statusDesc = "sending";
+		break;
+	case CRON_STATUS_CONNECTING:
+		statusDesc = "connecting";
+		break;
+	case CRON_STATUS_SUCCEEDED:
+		statusDesc = "succeeded";
+		break;
+	case CRON_STATUS_FAILED:
+		statusDesc = "failed";
+		break;
+	default:
+		break;
+	}
+	return statusDesc;
+}
+
+
+/*
+ * JobRunDetailsTableExists returns whether the job_run_details table exists.
+ */
+static bool
+JobRunDetailsTableExists(void)
+{
+	Oid cronSchemaId = get_namespace_oid(CRON_SCHEMA_NAME, false);
+	Oid jobRunDetailsTableOid = get_relname_relid(JOB_RUN_DETAILS_TABLE_NAME,
+												  cronSchemaId);
+
+	return jobRunDetailsTableOid != InvalidOid;
+}
